(* Timbuk - A Tree Automata Library (for Ocaml)
   Copyright (C) 2000-2001 Thomas Genet and Valrie Viet Triem Tong

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU Library General Public License version 2, as
   published by the Free Software Foundation.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

   See the GNU Library General Public License version 2 for more details
   (enclosed in the file LGPL).


   Version 1.1. Last modification 29/06/01
*)

open Common

module Completion
  (Alphabet_type: ALPHABET_TYPE)
  (Term: TERM_TYPE with type alphabet= Alphabet_type.t and type symbol= Alphabet_type.symbol)
  (TRS_type: TRS_TYPE with type alphabet= Alphabet_type.t and type term= Term.t)
  (State_set_type: STATE_SET_TYPE with type state= Term.t
				  and type alphabet= Alphabet_type.t)
  (Automata_type: AUTOMATA_TYPE with type substitution= Term.substitution
				and type symbol= Alphabet_type.symbol
				and type state= State_set_type.state
				and type state_set= State_set_type.t 
				and type alphabet= Alphabet_type.t 
				and type term= Term.term 
				and type transition_table= TRS_type.t 
				and type rule= TRS_type.rule)
  (Spec_type: AUTOMATON_SAVING_TYPE with type automaton= Automata_type.t)
  (Gamma: GAMMA_TYPE with type transition_table= TRS_type.t 
		     and type alphabet= Alphabet_type.t
		     and type symbol= Alphabet_type.symbol
		     and type state= State_set_type.state
		     and type transition= TRS_type.rule
		     and type ('a, 'b) folder= ('a, 'b) Automata_type.folder)=
struct
  module State_type= Term
  module Trans_type= TRS_type

  (*  channels for user interaction *)

  let out_chan= ref stdout
  let in_chan= ref stdin


  (* a flag for user interruption (Ctrl-C) *)

  let user_interrupt= ref false


  (* a reference on the current automaton to be completed, used to give the current
     state of the completion when the completion is interrupted thanks to a Ctrl-C *)

  let current_automaton= ref (Automata_type.make_automaton 
		      Alphabet_type.new_alphabet
		      Alphabet_type.new_alphabet
		      State_set_type.empty
		      State_set_type.empty
		      Trans_type.empty
		      Trans_type.empty)
		 
  let gamma= Gamma.gamma

  type gamma_content = Gamma.gamma_content
  type automaton = Automata_type.t
  type alphabet = Alphabet_type.t	
  type state_set = State_set_type.t
  type state = State_type.t
  type transition_table = Trans_type.t
  type transition = Trans_type.rule


    
  (* In a completion_state, we find the 
     - trs: current TRS (or part of it) used for
     completion, 
     - states: the completed set of states (with their description)
     - state_ops: the completed set of state_ops, 
     - delta: the current delta (folder of transitions categorized by symbol and then by state), 
     - g_content: which is the dynamic approximation information used by gamma,
     - new_trans: the new_transition found during a step of completion,
     can evenly be dupplicated or present in delta also, this has to be checked at then end of each completion step
     - index: the last natural used to label new states.
  *)

  type completion_state = 
      {trs: TRS_type.t; alphabet: alphabet; states:state_set; state_ops: alphabet; final:state_set;
       delta: (Alphabet_type.symbol, (state, transition list) Automata_type.folder) Automata_type.folder;
       g_content: gamma_content; new_trans: transition_table;
       index: int} 

  (* in and out channels are the same as those used in Gamma *)

  let in_chan= ref !Gamma.in_chan
  let out_chan= ref !Gamma.out_chan

  let get_trs (cs:completion_state)= cs.trs
  let get_alphabet (cs: completion_state)= cs.alphabet
  let get_states cs= cs.states
  let get_state_ops cs= cs.state_ops
  let get_transitions cs= cs.delta
  let get_gcontent (cs:completion_state)= cs.g_content
  let get_new_trans cs= cs.new_trans 
  let get_automaton_from_completion_step cs= Automata_type.make_automaton cs.alphabet
					       cs.state_ops cs.states cs.final 
					       (Gamma.get_prior cs.g_content)
					       (Trans_type.list_to_trs (Automata_type.bi_folder_flatten cs.delta))

  (* Construction of a completion state *)

  let make_completion_state trs aut gamma_content i=
    {trs= trs; alphabet= Automata_type.get_alphabet aut; states=Automata_type.get_states aut; 
     state_ops= Automata_type.get_state_ops aut; final= Automata_type.get_final_states aut;
     delta= Automata_type.transitions_by_state_by_symbol 
	      (Trans_type.to_list (Automata_type.get_transitions aut)); 
     g_content= Gamma.add_prior_to_gamma (Automata_type.get_prior aut) gamma_content; 
     new_trans=Trans_type.empty;
     index= i}

  (* Union of last completion step with the complete trs. Other informations such 
     as gamma_content etc... comes from the completion_state and are conserved in 
     the result of the union. the new_trans are added to the bifolder delta if they do not already exist *)

  let union trs last_cs=
    {trs=trs; alphabet= last_cs.alphabet; states= State_set_type.union (Automata_type.states_of_transitions last_cs.new_trans) last_cs.states; 
     state_ops=  last_cs.state_ops; 
     final= last_cs.final;
     delta= Automata_type.bi_folder_add_trans_list (Trans_type.to_list last_cs.new_trans) last_cs.delta;
     g_content= last_cs.g_content; new_trans=Trans_type.empty;
     index= last_cs.index}


  (* after one step of completion on the first rule, we take into account each new information and retrieve
     the first rule of the trs *)

  let modif_after_one_step cs new_trans new_gamma_content new_state_ops new_i=
    {trs= TRS_type.remainder cs.trs; alphabet= cs.alphabet; states= cs.states; state_ops= new_state_ops; 
     final= cs.final;
     delta= cs.delta; g_content= new_gamma_content; new_trans= Trans_type.union new_trans cs.new_trans;
     index= new_i}
    
    
  (* the lexer for adding rules interactively *)	

  let lexer = Genlex.make_lexer [",";"(";")";"->";"."]

		
  (* suppressing empty epsilon transitions and double occurrences of transitions in a trs
     coming from state-merging *) 
		
  let suppress_empty_trans a=
    let rec aux trans= 
      if Trans_type.is_empty trans then trans
      else 
	let t= Trans_type.first trans and lrem= Trans_type.remainder trans in
	let remainder= aux lrem in
	  if (Term.equal (Trans_type.lhs t) (Trans_type.rhs t)) then remainder
	  else Trans_type.add t remainder
    in Automata_type.make_automaton (Automata_type.get_alphabet a) (Automata_type.get_state_ops a) (Automata_type.get_states a) 
	 (Automata_type.get_final_states a) (aux (Automata_type.get_prior a)) (aux (Automata_type.get_transitions a))
	 

  (* parsing of merging rules *)

  let rec parse_merging_rules alphabet state_ops=
    let merging_rules_aux alphabet state_ops=
      parser
	  [< trs= Trans_type.parse_ground_special alphabet state_ops ;
	     'Genlex.Kwd "." >] -> trs 
    in 
    let rules= 
      try (merging_rules_aux alphabet state_ops (lexer (Stream.of_channel !in_chan)))
      with _ -> 
	begin
	  Printf.fprintf !out_chan "%s" "Syntax error, try again:\n"; 
	  flush !out_chan;
	  parse_merging_rules alphabet state_ops
	end
	
    in if (List.for_all (function t-> (Automata_type.is_state_config (Trans_type.lhs t)) &&
			   (Automata_type.is_state_config (Trans_type.rhs t))) (Trans_type.to_list rules))
      then 
	rules 
      else 
	begin
	  Printf.fprintf !out_chan "%s" "Be careful, one of the rules is not of the form q1 -> q2 where q1 and q2 are states\nTry again:\n"; 
	  flush !out_chan;
	  parse_merging_rules alphabet state_ops
	end
	
	
  (* state merging: merging together some states and renaming state label by another. *)
	
  let rec merge_state (cs: completion_state)=
    begin
      Printf.fprintf !out_chan "%s" "Give a sequence of rewrite rules on state labels ended by a dot: '.'\n\n";
      flush !out_chan;
      let transitions= parse_merging_rules cs.alphabet cs.state_ops in
      let trs_from_transitions= 
	TRS_type.list_to_trs
	  (List.map 
	     (function t-> 
		TRS_type.new_rule (Automata_type.state_label (Trans_type.lhs t))
		  (Automata_type.state_label (Trans_type.rhs t))) (Trans_type.to_list transitions)) in
      let (new_automaton:Automata_type.t)= suppress_empty_trans 
					     ((Automata_type.rewrite_state_labels 
						 (get_automaton_from_completion_step cs) trs_from_transitions):Automata_type.t) in
	make_completion_state cs.trs new_automaton (Gamma.set_prior_in_gamma Trans_type.empty cs.g_content) cs.index

    end
    

  (* Epures a list of transition with regards to a bifolder of transitions. Transitions to be retrieved are of
     the form:
     
     q -> q, 
     q1 -> q2, if all transitions of q1 are already (syntactically) in q2
     c -> q, if we already have c->*delta
  *)
    
  let rec epure new_trans delta=
    match new_trans with
	[] -> []
      | t::lrem -> 
	  let lh= Trans_type.lhs t in
	  let rh= Trans_type.rhs t in
	  let remainder= epure lrem delta in
	    if Automata_type.is_state_config lh then
	      if (Term.equal lh rh) || (Automata_type.is_covered lh rh delta)
	      then remainder
	      else t::remainder
	    else 
	      if Automata_type.is_recognized_into lh rh delta then remainder
	      else t::remainder
		

  let complet_all_trans_first_rule cs = 
    let rule= TRS_type.first cs.trs in
    let lhs= TRS_type.lhs rule in
    let rhs= TRS_type.rhs rule in
      
    let top_symb= Term.top_symbol lhs in
    let concerned_transitions= 
      try (Automata_type.folder_flatten 
             (Automata_type.folder_assoc top_symb cs.delta)) with 
	  Not_found -> [] in

    (* matching_solutions is a list of pairs (lsubst, q) where lsubst is a list of 
       substitutions s such that for all s: lhs s ->A q *) 
      
    let matching_solutions= 
      List.map (function trans -> 
		  (   (Automata_type.matching lhs (Trans_type.lhs trans) cs.delta),
		      Trans_type.rhs trans)) concerned_transitions in

    (* new_transitions is a list of new transitions corresponding to matching solutions: for all pair (lsubst, q)
       we add all transitions (rhs s) -> q  where s is a substition of lsubst *)
      
    let new_transitions= 
      List.flatten (List.map (function (lsubst, q) -> 
				let temp= Term.apply_several lsubst rhs in
				  List.map (function t -> Trans_type.new_rule t q) temp)
		      matching_solutions)

    in

    (* new transitions of the form q->q, and c -> t, such that we already have c->*D q are not
       necessary and do not need to be considered, where D=delta U new_trans *)

    let relevant_new_transitions= (Trans_type.list_to_trs (epure new_transitions (Automata_type.bi_folder_add_trans_list 
							   (Trans_type.to_list cs.new_trans) cs.delta))) in
      
    (* we apply the approximation gamma in order to normalize the transitions *)
      
    let (normalized_trans_with_gamma,
	 newgc, new_state_ops, new_i)= Gamma.gamma relevant_new_transitions 
					 cs.delta cs.g_content cs.alphabet 
					 cs.state_ops cs.index 

    in modif_after_one_step cs normalized_trans_with_gamma newgc new_state_ops new_i


  (* one step of completion: apply all rules on all transitions once *)

  let rec complet_one_step cs= 
    if TRS_type.is_empty cs.trs then cs
    else
      let one_step_first_rule= complet_all_trans_first_rule cs in
	complet_one_step one_step_first_rule



  (* Given an automaton a and a list of states q1,..., qn, this function verifies if 
     the intersection automaton a1= a /\ (a with q1 as final state) is empty, if not
     it verifies if intersection a2= a1 /\(a1 with q2 as final state) is empty, etc...*)

  let rec is_empty_multiple_intersection a a_init lstates=
    match lstates with
	[] -> false
      | q::lrem -> 
	  let intersec= 
	    Automata_type.simplify 
	      (Automata_type.inter a 
		 (Automata_type.modify_final a_init (State_set_type.list_to_set [q]))) in
	    if Automata_type.is_empty intersec then true
	    else is_empty_multiple_intersection intersec a_init lrem
	      

  (* is there a list l of states in l2 such that l<= l1, i.e. is there an empty
     intersection of states l such that the language recognized by l1 should be smaller
     or equal and thus also empty *)

  let covered intersec linter= 
    List.exists (function l -> Common.subeq_list l intersec) linter

  let rec subsume intersec l_intersec=
    match l_intersec with
	[] -> intersec::l_intersec
      | existing_intersec::lrem ->
	  (* if intersec subsumes one existing_intersec then we replace it *)
	  if Common.subeq_list intersec existing_intersec 
	  then intersec::lrem
	  else existing_intersec::(subsume intersec lrem)

	    
  (* (is_disjoint: completion_state -> (state list list) -> (state list list) -> (bool,(state list)))= 

     fun cs linter empty_inter-> *)

  let rec is_disjoint cs linter empty_inter=
    match linter with
	[] -> (true, [])
      | one_inter::lrem ->
	  if covered one_inter empty_inter
	  then is_disjoint cs lrem empty_inter
	  else 
	    begin
	      Printf.fprintf !out_chan "%s" ("Checking intersection: "^(String.concat " ^ " (List.map State_type.to_string one_inter)^" ..."));
	      flush !out_chan;
	      let a= (Automata_type.make_automaton cs.alphabet cs.state_ops
			cs.states (State_set_type.list_to_set [List.hd one_inter])
			(Gamma.get_prior (cs.g_content))
			(Trans_type.list_to_trs (Automata_type.bi_folder_flatten cs.delta))) in
		if (is_empty_multiple_intersection a a (List.tl one_inter))
		then 
		  begin
		    Printf.fprintf !out_chan "%s" " done.\n";
		    flush !out_chan;
		    is_disjoint cs lrem (subsume one_inter empty_inter)
		  end
		else (false, one_inter)
	    end


  let check_linearity cs=
    let disj_constraint= 
      Automata_type.disjointness_constraint (TRS_type.non_linear_lhs cs.trs) cs.delta
    in 
      is_disjoint cs disj_constraint []

  (* prints the result of intersection between current completed automaton and a list of automata given
     with their names *)

  let rec verif_inter aut_comp laut=
    match laut with
	[] -> ()
      | (name, a)::lrem -> 
	  let intersec= Automata_type.simplify (Automata_type.inter aut_comp a) in
	  let interstring= Automata_type.to_string intersec in
	    begin
	      Printf.fprintf !out_chan "%s" ("\nIntersection with "^name^" gives"^
					     (if Automata_type.is_empty intersec then " (the empty automaton)" else " (not empty)")^":\n\n"^interstring^"\n\nPress Enter to continue.");
	      flush !out_chan;
	      let _= input_line !in_chan in ();
		verif_inter aut_comp lrem
	    end
	    
	    
  (* user interaction capture for completion *)

  let rec answer()= 
    (Printf.fprintf !out_chan "%s" "(c/a/m/s/d/i/v/w/f/u/q)? ";
     flush !out_chan;
     let rep= input_line !in_chan in
       match rep with
	   "c"|"C"|"complete" -> 'c'
	 | "a"|"A"|"auto" -> 'a'
	 | "m"|"M"|"merge" -> 'm'
	 | "s"|"S"|"see" -> 's'
	 | "d"|"D"|"display" -> 'd'
	 | "i"|"I"|"inter" -> 'i'
	 | "v"|"V"|"verify" -> 'v'
	 | "f"|"F"|"forget" -> 'f'
	 | "u"|"U"|"undo" -> 'u'
	 | "q"|"Q"|"quit" -> 'q'
	 | "w"|"W"|"write" -> 'w'
	 | _ -> 
	     (Printf.fprintf !out_chan "%s" "Answer by ";
	      flush !out_chan;
	      answer()))

 (* user interaction capture for interruption handling *)

  let rec answer_interrupt()= 
    (Printf.fprintf !out_chan "%s" "(c/f/s/q)? ";
     flush !out_chan;
     let rep= input_line !in_chan in
       match rep with
	   "c"|"C"|"continue" -> 'c'
	 | "f"|"F"|"finish" -> 'f'
	 | "s"|"S"|"see" -> 's'
	 | "q"|"Q"|"quit" -> 'q'

	 | _ -> 
	     (Printf.fprintf !out_chan "%s" "Answer by ";
	      flush !out_chan;
	      answer_interrupt()))
(* Completion behaviour in case of user interruption (Ctrl-C).
   If Ctrl-C is hit, the current completion step is paused and the user
   is asked whether he wants 
       - to see current completed automaton (the one on which critical
         pairs are computed)
       - continue the current completion step
       - finish the current completion step and return to the menu
       - interrupt the current completion step NOW (thus loose current computation) and
         go back to the menu *)

  let rec interruption_handler (i:int)=
    begin
      Printf.fprintf !out_chan "%s" "\nDo you want to:\n\n(c)ontinue the completion\n(s)ee the current automaton (on which critical pairs are computed)\n(f)inish the current completion step and then go back to the menu\n(q)uit the completion step NOW (thus loose current computation) and go back to the menu\n\n";
      
      match answer_interrupt() with
	  'c' -> ()
	| 's' ->  
	    begin
	      Printf.fprintf !out_chan "%s"
		(Automata_type.to_string !current_automaton);
	      flush !out_chan;
	      interruption_handler i
	    end
	| 'f' ->
	    user_interrupt:= true
	| 'q' -> raise Sys.Break
	| _ -> raise Sys.Break
    end


  (* reading a file name *)

  let rec file_name()= 
    (Printf.fprintf !out_chan "%s" "Give a file name (give an empty file name to cancel): ";
     flush !out_chan;
     let rep= input_line !in_chan in
       if rep="" then (false, "") else (true, rep))


  (* Runs a completion on a given trs rws, a tree automaton a, an approximation material gamma_content.
     In the completion state: cs, trs may decrease during completion (for exploration purposes)
     delta can only increase with new transitions, new_delta contains all new transitions obtained by applying
     all rules on all transitions ONE time and then is added to delta.
     index increase during completion each time that we generate a new state automatically.
     states, state_ops, and gamma_content are intialised with the automaton and
     are only enriched during the completion from one step to another.

     laut is a list of pairs (string, tree automaton) on which user can try to intersec the current automaton computed during 
     completion *)

  let complet trs aut gamma_content laut=
    let rec menu lcs=
      try (
      begin
	Printf.fprintf !out_chan "%s" "\nDo you want to:\n\n(c)omplete one step  (use Ctrl-C to interrupt if necessary)\ncomplete (a)ll steps (use Ctrl-C to interrupt if necessary)\n(m)erge some states\n(s)ee current automaton\n(d)isplay the term rewriting system\n(i)ntersection with verif automata\n(v)erify linearity condition on current automaton\n(w)rite current automaton to disk\n(f)orget old completion steps\n(u)ndo last step\n(q)uit completion\n\n";
	match answer() with
	    'c' -> 
	      begin
		current_automaton:= (get_automaton_from_completion_step (List.hd lcs));
		complet_on_state lcs
	      end
	  | 'a' -> 
	      begin
		user_interrupt:= false;	
		(* Ctrl-C is mapped to a specific handler *)
		Sys.set_signal Sys.sigint (Sys.Signal_handle interruption_handler);
		current_automaton:= (get_automaton_from_completion_step (List.hd lcs));
		complete_all_steps lcs
	      end
	  | 'm' -> menu ((merge_state (List.hd lcs))::lcs)
	  | 's' -> 
	      begin
		Printf.fprintf !out_chan "%s"
		  (Automata_type.to_string (get_automaton_from_completion_step (List.hd lcs)));
		flush !out_chan;
		Printf.fprintf !out_chan "%s" "\n";
		menu lcs
	      end
	  | 'd' -> 
	      begin
		Printf.fprintf !out_chan "%s"
		  (TRS_type.to_string (get_trs (List.hd lcs)));
		Printf.fprintf !out_chan "%s" "\n";
		flush !out_chan;
		menu lcs
	      end
	      
	  | 'q' -> 
	      if not ((TRS_type.non_linear_lhs trs)=[]) then
		begin
		  Printf.fprintf !out_chan "%s" "Your TRS is non left-linear the completion is correct only if\nyou have verifed the linearity condition\nDo you want to verify it before quitting?";
		  flush !out_chan;
		  if Gamma.go() then menu lcs
		  else (get_automaton_from_completion_step (List.hd lcs))
		end
	      else
		(get_automaton_from_completion_step (List.hd lcs))
		
	  | 'f' -> 
	      if List.length lcs > 1 then 
		begin
		  Printf.fprintf !out_chan  "%s" "Forget old completion steps!\n";
		  menu [(List.hd lcs)]
		end
	      else 
		begin
		  Printf.fprintf !out_chan  "%s" "Nothing to forget!\n";
		  menu lcs
		end

	  | 'u' -> 
	      if List.length lcs > 1 then 
		begin
		  Printf.fprintf !out_chan  "%s" "Undo last completion step!\n";
		  current_automaton:= (get_automaton_from_completion_step (List.hd (List.tl lcs)));
		  menu (List.tl lcs)
		end
	      else 
		begin
		  Printf.fprintf !out_chan  "%s" "Nothing to undo!\n";
		  menu lcs
		end

	  | 'i' -> 
	      begin
		if laut=[]
		then 
		  begin
		    print_string "List of automata is empty: cannot compute intersections!\n\n";
		    menu lcs;
		  end
		else 
		  begin
		    verif_inter (get_automaton_from_completion_step (List.hd lcs)) laut;
		    menu lcs
		  end
	      end

	  | 'v' -> 
	      let (is_lin, lstates)= (check_linearity (List.hd lcs)) in
		if is_lin then
		  begin
		    Printf.fprintf !out_chan  "%s" "No linearity problem!\n";
		    menu lcs
		  end
		else 
		  begin 
		    Printf.fprintf !out_chan  "%s" "Linearity problem with states: ";
		    let _=List.map (function q -> Printf.fprintf !out_chan  "%s" ((Term.to_string q)^" ")) lstates in
		      menu lcs
		  end		 
	  | 'w' -> 
	      let (ok, file_name)= file_name() in
		if not ok then (menu lcs)
		else 
		  begin
		    (try (Spec_type.save_automaton (get_automaton_from_completion_step (List.hd lcs))
			    "current" file_name)
		     with
			 _ -> 
			   begin
			     Printf.fprintf !out_chan  "%s" ("Cannot write in file: "^file_name^"!\n");
			     flush !out_chan
			   end);
		    menu lcs
		  end

	  | _ -> failwith "Bug in parsing of user command: Completion.answer()"
      end)
      with
	  Sys.Break -> 
	    begin
	      Printf.fprintf !out_chan "%s" "\n\nLast completion step has been aborted!!\n";
	      Printf.fprintf !out_chan "%s" "---------------------------------------\n\n";
	      (* Ctrl-C is re-mapped to its default behaviour *)
	      Sys.set_signal Sys.sigint (Sys.Signal_default);
	      menu lcs
	    end

      (* complete only one step *)

    and complet_on_state lcs= 
      let one_step= (complet_one_step (List.hd lcs)) in
	if Trans_type.is_empty(get_new_trans one_step) 
	then 
	  begin 
	    Printf.fprintf !out_chan "%s" "\nAutomaton is complete!!\n";
	    Printf.fprintf !out_chan "%s" "------------------------\n\n"; 
	    menu ((union trs one_step)::lcs)
	  end
	else 
	  menu ((union trs one_step)::lcs)

    (* complete until the automaton is complete or user type Ctrl-C *)

    and complete_all_steps lcs=
	 try (
	   let one_step= (complet_one_step (List.hd lcs)) in
	     if Trans_type.is_empty(get_new_trans one_step)
	     then 
	       begin
		 Printf.fprintf !out_chan "%s" "\nAutomaton is complete!!\n";
		 Printf.fprintf !out_chan "%s" "------------------------\n\n"; 
		 menu ((union trs one_step)::lcs)
	       end
	     else
	       if !user_interrupt then
		 begin
		   (* Ctrl-C is re-mapped to its default behaviour *)
		   Sys.set_signal Sys.sigint (Sys.Signal_default);
		   menu ((union trs one_step)::lcs)
		 end
	       else
	       begin
		 let new_comp_state= (union trs one_step) in
		   current_automaton:= (get_automaton_from_completion_step new_comp_state);
		   complete_all_steps (new_comp_state::lcs)
	       end)
	 with
	     Sys.Break -> 
	       begin
		 Printf.fprintf !out_chan "%s" "\n\nLast completion step has been aborted!!\n";
		 Printf.fprintf !out_chan "%s" "---------------------------------------\n\n";
		 (* Ctrl-C is re-mapped to its default behaviour *)
		 Sys.set_signal Sys.sigint (Sys.Signal_default);
		 menu lcs
	       end

    in
      begin
	Printf.fprintf !out_chan "%s" "\nSTARTING COMPLETION...\n";
	Printf.fprintf !out_chan "%s" "-----------------------\n\n\n"; 
	flush !out_chan;
	current_automaton:= aut;
	menu [(make_completion_state trs aut gamma_content 0)]
      end




end
