(* Timbuk - A Tree Automata Library (for Ocaml)
   Copyright (C) 2000-2001 Thomas Genet and Valrie Viet Triem Tong

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU Library General Public License version 2, as
   published by the Free Software Foundation.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

   See the GNU Library General Public License version 2 for more details
   (enclosed in the file LGPL).


   Version 1.1. Last modification 29/06/01
*)

(* This is a sample main program to use with the Timbuk library. It simply reads 
   a spec file name given as a parameter of the program and starts a completion 
   with 
   
   - the first TRS of the spec,
   - the first tree automaton of the spec, 
   - the first approximation (if it exists) of the spec (or the approximation whose name
     is given with the option flag --approx
   - no approximation at all if the --noapprox flag is present.

   and it permits to compute intersections with the completed automaton and all the other
   automata of the specification file (with the (i) command of the completion mechanism.

   Examples:        main --approx Simple example_comp2.txt
                    main example_comp2.txt
                    main --noapprox example_comp2.txt
                    main example_comp.txt

*)

open Symbol
open Alphabet
open Variable
open Variable_set
open Term
open State_set
open State_content
open Rewrite
open Automaton
open Gamma
open Specification
open Completion

(* module instanciation *)

module MyF= Alphabet(Symbol)
module MyX= Variable_set(Variable)
module Myterm=Term(Symbol)(MyF)(Variable)(MyX)
module Mytrs= RewriteSystem(MyF)(MyX)(Myterm)
module MyQ= State_set(Symbol)(MyF)(Myterm)(State_content)
module Myautomaton= TreeAutomata(Symbol)(MyF)(Variable)(Myterm)(State_content)(Mytrs)(MyQ)
module MyGamma= Gamma(Symbol)(MyF)(MyX)(Myterm)(Mytrs)(MyQ)(Myautomaton)
module Myspec= Specification(MyF)(MyX)(Mytrs)(Myautomaton)(MyGamma)
module MyComp = Completion(MyF)(Myterm)(Mytrs)(MyQ)(Myautomaton)
		  (Myspec)(MyGamma)

let usage= ("Usage: "^Sys.argv.(0)^" [options] [specification file name]\nOptions: --approx A  where A is an approximation name of the specification\n                     (otherwise the first one is used)\n         --noaprox to ignore the approximations in the specification\n");;
let arg_number= Array.length Sys.argv;;
let approx= ref "";;

if arg_number<=1 or arg_number > 4 then print_string usage
else 
  begin
    if arg_number=4 then
      if Sys.argv.(1) = "--approx" then approx:= Sys.argv.(2)
      else print_string usage
    else 
      if arg_number=3 then
	if Sys.argv.(1) = "--noapprox" then approx := "##NOAPPROX##"
	else ();
    
    (* parsing of the specification file *)
    let s= Myspec.file_parse (Sys.argv.(arg_number-1)) in
      
    (* extraction of named trs, automata and approximations: i.e. lists of pairs
       (name, object) *)

    let ltrs= Myspec.get_list_trs s in
    let laut= Myspec.get_list_automata s in
    let lapprox= Myspec.get_list_approximation s in
      if ltrs=[] then print_string "Missing term rewriting system in the specification"
      else 
	if laut=[] then print_string "Missing tree automaton in the specification"
	else 
	  let (_, trs)=List.hd ltrs in
	  let (_, aut)=List.hd laut in
	  let app=
	    if not (!approx = "" or  !approx="##NOAPPROX##") then 
	      (* if an approximation has been given on the command line, we use it *)
	      Myspec.get_approximation !approx s
	    else 
	      if (lapprox=[] or !approx="##NOAPPROX##")
		(* if there is no approximation in the file or the --noapprox flag
		   has been used then we use the empty approximation *)
	      then MyGamma.empty
		(* we use the first approximation of the file *)
	      else ((function x,y -> y)(List.hd lapprox))
	  in
	  let aut_completed=  MyComp.complet trs aut app (List.tl laut) in
	    Myautomaton.print aut_completed
  end
  
